
#ifndef __POLYNOMIAL_CPP
    #define __POLYNOMIAL_CPP


    template<class T> template<class U> int polynomial<T>::signum(const U& x) const
    {
        if(x < (U)(0))
            return -1;
        if(x > (U)(0))
            return 1;
        return 0;
    }
    
    
    template<class T> polynomial<T>::polynomial()
    {
        coefficients=std::vector<T>(1,(T)(0));
    }

    template<class T> polynomial<T>::polynomial(T X)
    {
        coefficients=std::vector<T>(1,X);
    }
    
    template<class T> polynomial<T>::polynomial(size_t grade, T* coeffs)
    {
        coefficients = std::vector<T>(grade+1,(T)(0));
        if(coeffs != NULL)
        {
            for(size_t i = 0; i < grade; i++)
                coefficients[i] = coeffs[i];
            RemoveLeadingZeros(); // in case coeffs==NULL, we basically only allocate the RAM
                                  // RemoveLeadingZeros would immediately free it again...
        }
    }

    template<class T> polynomial<T>::polynomial(const std::vector<T>& coeffs) 
    {
        coefficients = coeffs;
        RemoveLeadingZeros();
    }

    template<class T> polynomial<T>::polynomial(const polynomial<T>& X)
    {
        coefficients = X.coefficients;
    }

    
    
    template<class T> void polynomial<T>::operator=(const polynomial<T>& a)
    {
        coefficients=a.coefficients;
    }

    template<class T> size_t polynomial<T>::Grade() const
    {
        return coefficients.size()-1;
    }

    template<class T> const T& polynomial<T>::operator[](size_t index) const 
    {
        return coefficients[index];
    }

    template<class T> T& polynomial<T>::operator[](size_t index)
    {
        return coefficients[index];
    }
    
    template<class T> template<class U> U polynomial<T>::operator()(const U& x) const
    {
        U result = (U)(0);
        T CurrentXPower = (T)(1);
        for(int i = 0; i <= Grade(); ++i)
        {
            result += CurrentXPower * coefficients[i];
            CurrentXPower *= x;
        }
        return result;
    }

    template<class T> void polynomial<T>::RemoveLeadingZeros()
    {
        while(coefficients[Grade()] == (T)(0))
            coefficients.pop_back();
    }
    
    

   // --------------------------------------------------------------- comparison

   
   
    template<class T> bool polynomial<T>::operator==(const polynomial<T>& a) const
    {
        return coefficients == a.coefficients;
    }

    template<class T> bool polynomial<T>::operator!=(const polynomial<T>& a) const 
    {
        return coefficients != a.coefficients;
    }

    template<class T> bool polynomial<T>::operator<(const polynomial<T>& a) const 
    {
        if(Grade() < a.Grade())
            return true;
        if(Grade() > a.Grade())
            return false;

        for(int i = Grade(); i>=0; --i)
        {
            if(coefficients[i] < a[i])
                return true;
            if(coefficients[i] > a[i])
                return false;
        }
        return false;
    }

    template<class T> bool polynomial<T>::operator<=(const polynomial<T>& a) const
    {
        return operator<(a) || operator==(a);
    }

    template<class T> bool polynomial<T>::operator>(const polynomial<T>& a) const
    {
        return ! operator<=(a);
    }

    template<class T> bool polynomial<T>::operator>=(const polynomial<T>& a) const
    {
        return ! operator<(a);
    }


    
    // ---------------------------------------------------------------- operators


    
    template<class T> polynomial<T> polynomial<T>::operator+(const polynomial<T>& a) const
    {
        int max = (Grade() > a.Grade() ? Grade() : a.Grade());
        int min = (Grade() < a.Grade() ? Grade() : a.Grade());
        polynomial<T> result(max, NULL);

        int i = 0;
        for(; i <= min; ++i)
            result[i] = coefficients[i] + a[i];
        for(; i <= Grade(); ++i)
            result[i] = coefficients[i];
        for(; i <= a.Grade(); ++i)
            result[i] = a[i];

        result.RemoveLeadingZeros();
        return result;
    }

    template<class T> polynomial<T> polynomial<T>::operator-(const polynomial<T>& a) const 
    {
        int max = (Grade() > a.Grade() ? Grade() : a.Grade());
        int min = (Grade() < a.Grade() ? Grade() : a.Grade());
        polynomial<T> result(max, NULL);

        int i = 0;
        for(; i <= min; ++i)
            result[i] = coefficients[i] - a[i];
        for(; i <= Grade(); ++i)
            result[i] = coefficients[i];
        for(; i <= a.Grade(); ++i)
            result[i] = (T)(0) - a[i];

        result.RemoveLeadingZeros();
        return result;
    }

#include<iostream>    
    
    template<class T> polynomial<T> polynomial<T>::operator*(const polynomial<T>& a) const
    {
        size_t ResultGrade = Grade() + a.Grade();
        polynomial<T> result(ResultGrade, NULL);

        for(int i = 0; i <= ResultGrade; ++i) 
        {
            std::cout << "\ncoefficient " << i << ": 0";
            int UpperBound = std::min((int)Grade(), i);
            result[i] = (T)(0);
            for(int k = std::max(i-(int)a.Grade(),0);  k <= UpperBound;  ++k)
            {
                std::cout << "  +coeff[" << k <<"]*a.coeff[" << (i-k) << "]";
                result[i] += coefficients[k] * a[i-k];
            }
            std::cout<< "    == " << result[i] << "\n";
        }

//        result.RemoveLeadingZeros();
        return result;
    }

    template<class T> polynomial<T> polynomial<T>::operator/(const polynomial<T>& a) const
    {
        if(Grade() < a.Grade())
            return polynomial<T>((T)(0));
        polynomial<T> div(Grade()-a.Grade(),NULL);
        for(int i = 0; i <= div.Grade(); i++)
            div[i]=0;
        polynomial<T> mod(*this);

        for(int i = Grade(); i >= a.Grade(); --i)
        {
            div[i-a.Grade()] = mod[i] / a[a.Grade()];
            for(int k = a.Grade(); k >= 0; --k)
                mod[k+i-a.Grade()] -= a[k] * div[i-a.Grade()];
        }

        div.RemoveLeadingZeros();
        return div;
    }

    template<class T> polynomial<T> polynomial<T>::operator%(const polynomial<T>& a) const
    {
        // same code as operator / only here the rest of the division (mod)
        //  is returned instead of the division result (div)
        if(Grade() < a.Grade())
            return polynomial<T>(*this);
        polynomial<T> div(Grade()-a.Grade()+1,NULL);
        for(int i = 0; i <= div.Grade(); i++)
            div[i]=0;
        polynomial<T> mod(*this);

        for(int i = Grade(); i >= a.Grade(); --i)
        {
            div[i-a.Grade()] = mod[i] / a[a.Grade()];
            for(int k = a.Grade(); k >= 0; --k)
                mod[k+i-a.Grade()] -= a[k] * div[i-a.Grade()];
        }

        mod.RemoveLeadingZeros();
        return mod;
    }


    
    // ----------------------------------------------------- assignment operators

    

    template<class T> void polynomial<T>::operator+=(const polynomial<T>& X)
    {
        operator=(operator+(X));
    }

    template<class T> void polynomial<T>::operator-=(const polynomial<T>& X)
    {
        operator=(operator-(X));
    }

    template<class T> void polynomial<T>::operator*=(const polynomial<T>& X)
    {
        operator=(operator*(X));
    }

    template<class T> void polynomial<T>::operator/=(const polynomial<T>& X)
    {
        operator=(operator/(X));
    }

    template<class T> void polynomial<T>::operator%=(const polynomial<T>& X)
    {
        operator=(operator%(X));
    }

    

    // ---------------------------------------------- polynomial specific methods

    

    template<class T> polynomial<T> polynomial<T>::Derivation() const 
    {
        if(Grade() < 1)
            return polynomial<T>();

        polynomial<T> result(Grade()-1, NULL);
        for(int i = Grade(); i > 0; --i)
            result[i-1] = (T)(i) * coefficients[i];

        return result;
    }

    template<class T> polynomial<T> polynomial<T>::Integral() const
    {
        polynomial<T> result(Grade()+1, NULL);
        for(int i = Grade()+1; i > 0; --i)
            result[i] = coefficients[i-1]/(T)(i);
        result[0]=0;
        return result;
    }

    template<class T> polynomial<T> polynomial<T>::Normalize() const
    {
        T d = coefficients[Grade()];
        polynomial<T> result(Grade(), NULL);
        for(int i = Grade(); i >= 0; --i)
            result[i] = coefficients[i] / d;
        return result;
    }

    template<class T> bool polynomial<T>::Normal() const
    {
        return coefficients[Grade()] == (T)(1);
    }

    
    
    // ------------------------------------------------------------------------- find the zeroes (numerically)

    
    
    template<class T> template<class U> U polynomial<T>::ApproximateZero(U lower, U upper) const
    {
        if(lower > upper)
            return ApproximateZero(upper, lower);
        
        U middle = (lower+upper)/2;
        U lowerval = operator()(lower);
        U upperval = operator()(upper);
        U middleval = operator()(middle);
        
        if(signum(upperval) == signum(lowerval))
            throw "ApproximateZero: lower and upper values must have different signs";
        
        while(upper-lower > (U)(0.000001))
        {
            if( signum(lowerval) == signum(middleval) )
            {
                lower = middle;
                lowerval = middleval;
            }
            else if(signum(middleval) == signum(upperval))
            {
                upper = middle;
                upperval = middleval;
            }
            middle = (lower+upper)/(U)(2);
            middleval = operator()(middle);
        }

        // pick the element of {lower,middle,upper} with the smallest absolute value
        if(middleval * signum(middleval) > lowerval * signum(lowerval))
        {
            middle = lower;
            middleval = lowerval;
        }
        if(middleval * signum(middleval) > upperval * signum(upperval))
            middle = upper;
        return middle;
    }

    
    
    template<class T> template<class U> U polynomial<T>::FindZero(U start, U direction) const
    {
        U startval = operator()(start);
        U nextval = operator()(start+direction);
        
        while(signum(startval) == signum(nextval))
        {
            start += direction;
            direction *= (U)(2);
            startval = nextval;
            nextval = operator()(start+direction);
        }
        
        return ApproximateZero(start, start+direction);
    }
    
#include<iostream>
    
    template<class T> template<class U> std::vector<U> polynomial<T>::Zeroes() const
    {
        std::vector<U> result;
        if(Grade() <= 0) // in case grade == 0, there either is no zero or it is constantly zero
        {
            if(coefficients[0] == (T)(0))
                throw "the polynomial is constant 0. This cannot be modelled with an std::vector";
            // else, the empty result is correct
        }
        else if(Grade() == 1)
        {
            result.push_back( (U)(-coefficients[0] / coefficients[1]) );
        }
        else // >= 2
        {
            std::vector<U> DerivationZeroes = Derivation().Zeroes<U>();        // zeroes of the derivation


            typename std::vector<U>::iterator i = DerivationZeroes.begin();
            U ival = operator()(i != DerivationZeroes.end() ? *i : (U)(0));
            
            int sign_lim_neg = signum(coefficients[Grade()]);
            if(Grade() % 2 == 1)
                sign_lim_neg *= -1;
            
            
            if(sign_lim_neg != signum(ival))
                // there is a zero between -infinity and the first zero of the derivation
                result.push_back(FindZero(i != DerivationZeroes.end() ? *i : (U)(0), (U)(-1)));


            typename std::vector<U>::iterator i2 = i;
            if(i2 != DerivationZeroes.end())
                i2++;
            U i2val;
            while(i2 != DerivationZeroes.end())
            {
                i2val = operator()(*i2);
                
                if(signum(ival) != signum(i2val))
                    // there is a zero between these 2 zeros of the derivation
                    result.push_back(ApproximateZero(*i, *i2));
                    
                ival = i2val;
                i = i2;
                i2++;
            }
            
            
            if(signum(ival) != signum(coefficients[Grade()]))
                // there is a zero between the last zero and +infinity
                result.push_back(FindZero(i != DerivationZeroes.end() ? *i : (U)(0), (U)(1)));
        }
        return result;
    }
    
    

    // ------------------------------------------------------------------------- output to a stream like cout or an ofstream

    

    template<class X> std::ostream& operator << (std::ostream& os, const polynomial<X>& Poly) 
    {
        os << Poly.Grade() << " ";
        for(typename std::vector<X>::const_reverse_iterator i = Poly.coefficients.rbegin(); i != Poly.coefficients.rend(); i++)
            os << *i << " ";
        return os;
    }

    template<class X> std::istream& operator >> (std::istream& is, polynomial<X>& Poly)
    {
        size_t grade;
        is >> grade;
        X current;
        Poly.coefficients.clear();
        for(size_t i = grade+1; i > 0; --i)
        {
            is >> current;
            Poly.coefficients.push_front(current);
        }
        return is;
    }

    template<class T> std::string polynomial<T>::LaTeX(const char* OpeningBracket, const char* ClosingBracket) const
    {
        std::stringstream buffer(std::stringstream::out);
        bool first = true;
        for(int i = Grade(); i >= 0; --i)
        {
            if(coefficients[i] != (T)(0))
            {
                if(first)
                {
                    buffer << OpeningBracket << coefficients[i] << ClosingBracket;
                    first = false;
                }
                else
                {            
                    buffer << OpeningBracket << (coefficients[i] > (T)(0) ? " + " : " - ") << (coefficients[i]*signum(coefficients[i])) << ClosingBracket;
                }
                
                if(i >= 1)
                    buffer << "x";
                if(i > 1)
                    buffer << (i > 9 ? "^{" : "^") << i << (i>9?"}":"");
            }
        }
        return buffer.str();
    }

#endif
